#include <errno.h>

#include <vos/vutil/log.hh>
#include <vos/vutil/getaddrinfowrapper.hh>

#include <vos/vip/socketmultiplexer.hh>
#include <vos/vip/socketthread.hh>

using namespace VUtil;
using namespace VIP;

SocketMultiplexer::SocketMultiplexer(short int listenport)
{
    skt = socket(PF_INET, SOCK_DGRAM, 0);

    struct sockaddr_in sa;
    memset(&sa, sizeof(sockaddr_in), 0);
    sa.sin_family = PF_INET;
    sa.sin_port = htons(listenport);
    sa.sin_addr.s_addr = htonl(INADDR_ANY);
    if(bind(skt, (sockaddr*)&sa, sizeof(sockaddr_in)) == -1) {
        LOG("VIP", 3, "Couldn't bind socket to port " << listenport << ": "
            << strerror(errno));
        throw PortBindingError(strerror(errno));
    }

    SocketProcessorThread sp(this);

    new boost::thread(sp);
}

vRef<Connection> SocketMultiplexer::makeConnection(const std::string& hostname,
                                                   uint16_t port)
{
    struct addrinfo hints;
    struct addrinfo *res;

    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_DGRAM;

    int err;
    if((err = getaddrinfo_wrapper(hostname.c_str(), 0, &hints, &res)) != 0) {
        throw HostnameError(std::string("Couldn't look up hostname: ")
                                 + std::string(gai_strerror_wrapper(err)));
    }
    ((sockaddr_in*)res->ai_addr)->sin_port = htons(port);

    vRef<Connection> nc(new Connection(skt,
                                       (sockaddr_in*)res->ai_addr,
                                       this),
                        false);

    {
        boost::mutex::scoped_lock lk(connections_mutex);
        connections[(sockaddr_in*)res->ai_addr] = nc;
    }

    freeaddrinfo_wrapper(res);

    nc->sendSYN();

    if(nc->isConnected()) {
        callback->notifyNewConnection(nc, false);
    } else {
        LOG("VIP", 4, skt << ": Timed out trying to connect");
        return vRef<Connection>();
    }

    //std::cout << skt << ": Starting with seq num " << nc->lastRecvNum << " "
    //<< nc->sentSeqCounter << std::endl;

    return nc;
}

void SocketMultiplexer::setupNewConnection(uint8_t* buf, int sz,
                                           struct sockaddr_in* from)
{
    vRef<Connection> nc(new Connection(skt, from, this), false);

    LOG("VIP", 6, "Adding to connections set " << nc->getRefCount());

    {
        boost::mutex::scoped_lock lk(connections_mutex);
        connections[from] = nc;
    }

    LOG("VIP", 5, skt << ": Sending SYN reply");

    nc->handleSYN(buf, sz);

    callback->notifyNewConnection(nc, true);
}


void SocketMultiplexer::handlePacket(uint8_t* buf, int ret, sockaddr_in* from)
{
    LOG("VIP", 5, skt << ": Got packet of length " << ret);

    vRef<Connection> conn;
    {
        boost::mutex::scoped_lock lk(connections_mutex);

        if(connections.count(from))
            conn = connections[from];
    }

    LOG("VIP", 5, skt << ": first num is " << (int)buf[0]);

    if(conn.isValid()) {
        switch(buf[0]) {
        case VIP_DATA:
        case VIP_PING:
            conn->splitIntoChunks(buf, ret);
            break;

        case VIP_SYN:
            /* This might mean that our SYN reply to initiate the
               connection was dropped, so the peer has done a SYN
               retransmit.  In this case, the SYN numbers to
               initiate the connection will be identical to the
               current state, and we reply with current sequence
               numbers.  If the SYN does not match the current
               state, then reset the connection.
            */
            if(!conn->handleSYN(buf, ret)) {
                removeConnection(from);

                LOG("VIP", 4, "resetting connection " << conn);
                conn->close(false, false, true);

                LOG("VIP", 5, "conn state is " << conn->getState());

                setupNewConnection(buf, ret, from);

                LOG("VIP", 5, "conn state *NOW* is " << conn->getState());
            }
            break;

        case VIP_RST:
            if(conn->getState() != Connection::Opening) {
                LOG("VIP", 6, "got RST, conn count is " << conn->getRefCount());
                removeConnection(from);
                LOG("VIP", 6, "got RST, conn count now is " << conn->getRefCount());
                conn->close(false, false);
            }
            break;
        }
    } else {
        switch(buf[0]) {
        case VIP_DATA:
        case VIP_PING:
        {
            unsigned char rst = VIP_RST;

            LOG("VIP", 1, "Got unknown data/ping pkt, sending RST pkt");

            sendto(skt, (char*)&rst, 1,
                   0, (struct sockaddr*)from,
                   (socklen_t)sizeof(sockaddr_in));
        }
        break;

        case VIP_SYN:
            /* Someone wants to initiate a connection, so set
               up a connection structure and reply with a
               matching SYN.
            */
            setupNewConnection(buf, ret, from);
            break;

        case VIP_RST:
            break;
        }
    }
}

void SocketMultiplexer::removeConnection(sockaddr_in* from)
{
    boost::mutex::scoped_lock lk(connections_mutex);
    connections.erase(from);
}
